# Copyright 2012© MaestroDev.  All rights reserved.
require 'onstomp'
require 'onstomp/failover'

module Maestro
  module MQHelper
    class MQError < StandardError
    end

    @@INFINITE_RETRIES = 2**32-1
    @@stomp = nil
    @@subs = {}
    @@tries = 0
    @@sleep = 20
    @@uri = nil

    def MQHelper.reset
      Maestro.log.info "Resetting MQHelper"
      @@stomp = nil
      @@subs = {}
      @@tries = 0
      @@sleep = 20
      @@uri = nil
    end

    def MQHelper.connection
      if @@stomp.nil?
        Maestro.log.warn "Request for stomp connection, but not set up yet!"
      end
      @@stomp
    end

    # only for testing
    def MQHelper.connection=(stomp)
      @@stomp = stomp
    end

    def MQHelper.connect
      if @@stomp.nil?
        Maestro.log.info "Request for stomp connection - creating"

        # retry info - the retryable below doesn't support 'unlimited' but if we give it a really big number...
        if @@tries == 0
          @@tries = Maestro.stomp_config['max_retries']

          if @@tries.nil?
            @@tries = @@INFINITE_RETRIES
            Maestro.log.info "Connect retry attempt count not specified, using ~INFINITE (really big number)"
          else
            Maestro.log.info "Connect retry attempt count not specified, using config file setting #{@@tries}"
          end
        end

        @@sleep = Maestro.stomp_config['retry_seconds'] || 20

        # New form is config.url = some_url
        # Old form is config.user, config.passcode, config.host, config.port,
        # config.ssl
        # Check for new config first, it overrides
        url = Maestro.stomp_config['url']

        if url && !url.empty?
          # New form
          is_failover = url.start_with?('failover:')
          options = {}

          if url.include?('+ssl')
            options[:ssl] = { :ca_file => Maestro.stomp_config['cert'] || nil, :verify_mode => OpenSSL::SSL::VERIFY_NONE }
          end

          @@uri = url

          if is_failover
            Maestro.log.info "Using OnStomp Failover Client"
            client = OnStomp::Failover::Client.new(url, options)
          else
            client = OnStomp::Client.new(url, options)
          end
        else
          # Old form
          Maestro.log.debug "Old-form stomp config being used"
          user = Maestro.stomp_config['user']
          passcode = Maestro.stomp_config['passcode']
          host = Maestro.stomp_config['host']
          port = Maestro.stomp_config['port'].to_s
          ssl = Maestro.stomp_config['ssl'] ? "+ssl" : ""
          cert = Maestro.stomp_config['cert'] || nil
  
          # construct the connection URI
          user_and_password = [user,passcode].reject{ |e| e.nil? || e.empty? }.join(":")
          host_and_port     = [host,port].reject{|e| e.nil? || e.empty?}.join(":")
          protocol          = ['stomp', ssl, '://'].reject{|e| e.nil? || e.empty?}.join
          @@uri             = protocol + [host_and_port, user_and_password].reject{|e| e.nil? || e.empty?}.reverse.join("@")
  
          if (!ssl.nil? && ssl == "+ssl")
            client = OnStomp::Client.new(@@uri, :ssl => {:ca_file => cert, :verify_mode => OpenSSL::SSL::VERIFY_NONE})
          else
            client = OnStomp::Client.new(@@uri)
          end
        end

        @@stomp = client
        reconnect
        raise "Failed To Connect To Maestro" unless client.connected?
      end
      
      @@stomp
    end

    def MQHelper.reconnect
      if !@@stomp
        raise Maestro::MQHelper::MQError, "Cannot reconnect, no existing MQ instance.  Call 'connect' to create initial connection, then use 'reconnect' if connection fails"
      end

      if !@@stomp.connected?
        Maestro::Utils.retryable(:tries => @@tries, :sleep => @@sleep) do
          Maestro.log.info "Connecting To stomp At #{@@uri}"
          @@stomp.connect
        end

        if @@stomp.connected?
          oldsubs = @@subs.dup
          @@subs.clear
          oldsubs.each do |dest, info|
            Maestro.log.info "Resubscribing to #{dest}"
            do_subscribe(dest, info['headers'], &info['callback'])
          end
        else
          Maestro.log.warn "Unable to reconnect to subscriptions because no MQ Connection"
        end
      end

      @@stomp.connected?
    end

    def MQHelper.disconnect
      if !@@stomp.nil?
        @@stomp.disconnect
      end
    end

    def MQHelper.subscribe(dest, headers={}, &cb)
      Maestro.log.info "MQHelper: Subscribing to #{dest}"
      do_subscribe(dest, headers, &cb)
    end

    def MQHelper.unsubscribe(dest, &cb)
      if @@subs.has_key?(dest)
        Maestro.log.info "MQHelper: Unsubscribing from #{dest}"
        @@stomp.unsubscribe(@@subs[dest]['frame'])
        @@subs.delete(dest)
      end
    end

    private

    def MQHelper.do_subscribe(dest, headers, &cb)
      sub = @@stomp.subscribe(dest, headers, &cb)
      @@subs[dest] = { 'frame' => sub, 'callback' => cb, 'headers' => headers }
      sub
    end
  end
end
